import numpy as np
import pandas as pd
import wuml
import os 
import sys

if os.path.exists('/home/chieh/code/wPlotLib'):
	sys.path.insert(0,'/home/chieh/code/wPlotLib')

import wplotlib
from sklearn.cluster import SpectralClustering
from sklearn.cluster import KMeans
from sklearn.mixture import GaussianMixture
from sklearn.cluster import DBSCAN
from sklearn.cluster import AgglomerativeClustering

class clustering:
	'''
	Automatically run clustering on data

	data: can be any data type
	method='Spectral Clustering', 'KMeans', 'GMM', 'Agglomerative'
	gamma= only used for Spectral clustering, default to 1
	'''

	def __init__(self, data, n_clusters, y=None, y_column_name=None, method='Spectral Clustering', kernel=None, gamma=1):
		NP = wuml.ensure_numpy
		S = np.squeeze

		self.X = X = NP(data)
		if y is not None:
			y = S(NP(y))
		elif y_column_name is not None:
			y = data[y_column_name].values
		elif type(data).__name__ == 'wData':
			try: y = data.Y
			except: pass

		if method == 'Spectral Clustering':
			model = SpectralClustering(n_clusters=n_clusters,assign_labels='discretize', random_state=0, gamma=gamma)
		elif method == 'KMeans':
			model = KMeans(n_clusters=n_clusters, random_state=0)
		elif method == 'GMM':
			model = GaussianMixture(n_components=n_clusters, random_state=0)
			model.fit(X)
			self.labels = model.predict(X)
		elif method == 'Agglomerative':
			model= AgglomerativeClustering(n_clusters=n_clusters)
		else: raise ValueError('Unrecognized Clustering Method')


		try:
			model.fit(X)
			self.labels = model.labels_
		except: pass

		self.method = method
		self.y = y
		self.n_clusters = n_clusters
		self.data = data

	def __str__(self):
		return str(self.labels)

	def plot_scatter_result(self, dimension_reduction_method='all'):
		methods = ['PCA', 'TSNE', 'KPCA']
		if dimension_reduction_method == 'all':
			for i, m in enumerate(methods):
				title = 'Using %s'%(self.method)
				imgText = '%d clusters\n%s'%(self.n_clusters, m)
				ϰ = wuml.dimension_reduction(self.X, method=m)
				if i == 0:
					plts = wplotlib.plot_clusters(ϰ, self.labels, title=title, subplot=131+i, figsize=(10,3), imgText=imgText)
				else:
					plts = wplotlib.plot_clusters(ϰ, self.labels, title=title, subplot=131+i, imgText=imgText)
				
			plts.show()
		else:
			title = '%d Clusters via %s'%(self.n_clusters, dimension_reduction_method)
			ϰ = wuml.dimension_reduction(self.X, method=dimension_reduction_method)
			wplotlib.plot_clusters(ϰ, self.labels, title=title)


	def nmi_against_true_label(self, print_out=True):
		NPR = np.round
		if self.y is None: raise Warning("warning: y is not defined!!")

		nmi = wuml.NMI(self.y, self.labels)

		column_names = ['Method', 'NMI']
		data = np.array([[self.method, nmi]])

		df = pd.DataFrame(data, columns=column_names,index=[''])
		if print_out: print(df)
		self.results = df
		return df


	def save_clustering_result_to_csv(self, path, with_data=False):
		if with_data:
			X = wuml.ensure_wData(self.data)
			X.append_columns(self.labels, column_names='cluster_labels')
			X.to_csv(path, float_format='%d')
		else:
			wuml.csv_out(self.labels, path, float_format='%d')


def run_every_clustering_algorithm(data, n_clusters, gamma=1):
	X = wuml.ensure_numpy(data)
	c_methods= ['Spectral Clustering', 'KMeans', 'GMM', 'Agglomerative']

	results = {}
	for C in c_methods:
		results[C] = clustering(data, n_clusters, method=C, gamma=gamma)
		results[C].plot_scatter_result()


#	methods = ['PCA', 'TSNE', 'KPCA']
#	for i, C in enumerate(c_methods):
#
#
#		for j, m in enumerate(methods):
#			title = 'Using %s'%(C)
#			imgText = '%d clusters\n%s'%(n_clusters, m)
#			ϰ = wuml.dimension_reduction(X, method=m)
#			if i == 0 and j==0:
#				plts = wplotlib.plot_clusters(ϰ, results[C].labels, title=title, subplot=431+i+j, figsize=(10,10), imgText=imgText)
#			else:
#				plts = wplotlib.plot_clusters(ϰ, results[C].labels, title=title, subplot=431+i+j, imgText=imgText)
#			print(i, j, 431+i*3+j)
#
#	plts.show()




#!/usr/bin/env python

import numpy as np
import wuml
from sklearn.model_selection import KFold

# data_list is a list of data, example: [X, Y] , or just [X]
def run_K_fold_on_function(K, data_list, func, func_args): 
	X = data_list[0]
	Xnp = wuml.ensure_numpy(X)
	all_results = []

	kf = KFold(n_splits=K, shuffle=True)
	kf.get_n_splits(Xnp)

	fold_id = 1
	data_one_fold_list = []
	for train_index, test_index in kf.split(Xnp):
		for dat in data_list:
			datNP = wuml.ensure_numpy(dat)
			data_one_fold_list.append([datNP[train_index], datNP[test_index]])

		func_args['one_fold_data_list'] = data_one_fold_list
		func_args['fold_id'] = fold_id
		fold_id += 1
		single_result = func(**func_args)
		all_results.append(single_result)

	return all_results




from sklearn import preprocessing
from wuml.type_check import *
import sys
import wuml
import pandas as pd
import torch
import numpy as np
from sklearn.preprocessing import LabelEncoder
from torch.autograd import Variable
from torch.utils.data import Dataset, DataLoader

class wData:
	def __init__(self, xpath=None, ypath=None, column_names=None, label_column_name=None, dataFrame=None, 
					X_npArray=None, Y_npArray=None, first_row_is_label=False, row_id_with_label=None, sample_id_included=False, 
					label_type=None, #it should be either 'continuous' or 'discrete'
					torchDataType=torch.FloatTensor, batch_size=20, columns_to_ignore=None,
					replace_this_entry_with_nan=None, preprocess_data=None):
		'''
			first_row_is_label :  True of False
			row_id_with_label: None, if the label is not the first row 0, set with this
			dataFrame: if dataFrame is set, it ignores the path and use the dataFrame directly as the data itself
			ypath: loads the data as the label
			label_column_name: if the label is loaded together with xpath, this separates label into Y
			preprocess_data: 'center and scale', 'linearly between 0 and 1', 'between 0 and 1 via cdf'
		'''
		self.label_column_name = label_column_name
	
		if dataFrame is not None:
			self.df = dataFrame
		elif X_npArray is not None:
			self.df = pd.DataFrame(X_npArray, columns=column_names)
		else:
			if first_row_is_label: 
				self.df = pd.read_csv (xpath, header=0, index_col=False)
			else:
				self.df = pd.read_csv (xpath, header=row_id_with_label, index_col=False)

		if replace_this_entry_with_nan is not None:
			self.df = self.df.replace(replace_this_entry_with_nan, np.nan)

		self.strip_white_space_from_column_names()

		self.columns = self.df.columns
		self.Y = None
		if Y_npArray is not None:
			self.Y = Y_npArray
		elif ypath is not None: 
			if label_type is None: raise ValueError('If you are using labels, you must include the argument label_type= "continuout" or "discrete"')
			self.Y = np.loadtxt(ypath, delimiter=',', dtype=np.float32)			
			if label_type == 'discrete': self.Y = LabelEncoder().fit_transform(self.Y)	#Make sure label start from 0
		elif label_column_name is not None:
			if label_type is None: raise ValueError('If you are using labels, you must include the argument label_type= "continuout" or "discrete"')
			self.Y = self.df[label_column_name].values
			if label_type == 'discrete': self.Y = LabelEncoder().fit_transform(self.Y)	#Make sure label start from 0
			self.delete_column(label_column_name)

		if columns_to_ignore is not None: self.delete_column(columns_to_ignore)

		self.X = self.df.values
		self.batch_size = batch_size
		self.shape = self.df.shape
		self.torchDataType = torchDataType				
		self.torchloader = None
		self.label_type = label_type

		if preprocess_data == 'center and scale':
			self.X = preprocessing.scale(self.X)
			self.df = pd.DataFrame(data=self.X, columns=self.df.columns)
		elif preprocess_data == 'linearly between 0 and 1':
			self.X = wuml.wuml.map_data_between_0_and_1(self.X, output_type_name='ndarray', map_type='linear')
			self.df = pd.DataFrame(data=self.X, columns=self.df.columns)
		elif preprocess_data == 'between 0 and 1 via cdf':
			self.X = wuml.wuml.map_data_between_0_and_1(self.X, output_type_name='ndarray', map_type='cdf')
			self.df = pd.DataFrame(data=self.X, columns=self.df.columns)


		if torch.cuda.is_available(): self.device = 'cuda'
		else: self.device = 'cpu'

	def strip_white_space_from_column_names(self):
		# make sure to strip white space from column names
		update_col_names = []
		for i, col_name in enumerate(self.df.columns):
			if type(self.df.columns[i]).__name__ == 'str':
				update_col_names.append(col_name.strip())
			else:
				update_col_names.append(col_name)

		self.df.columns = update_col_names

	def sort_by(self, column_name, ascending=True):
		self.df = self.df.sort_values(column_name, ascending=ascending)
		self.update_DataFrame(self.df)

	def get_columns(self, columns):
		columns = ensure_list(columns)
		subColumns = self.df[columns]
		return ensure_wData(subColumns)

	def update_DataFrame(self, df):
		self.df = df
		self.columns = self.df.columns
		self.X = self.df.values
		self.shape = self.df.shape
	
	def reset_index(self):
		self.df.reset_index(drop=True, inplace=True)	
		self.update_DataFrame(self.df)

	def append_rows(self, new_data):
		df = ensure_DataFrame(new_data, columns=self.columns)
		self.update_DataFrame(pd.concat([self.df,df], axis=0))

	def append_columns(self, new_data, column_names=None):
		df = ensure_DataFrame(new_data, columns=column_names)
		self.update_DataFrame(pd.concat([self.df,df], axis=1))

	def rename_columns(self, column_names):
		if type(column_names).__name__ == 'str':
			column_names = [column_names]

		self.df.columns = column_names
		self.columns = column_names

	def delete_column(self, column_name):
		if type(column_name) == type([]):
			for name in column_name:
				if name in self.df.columns:
					del self.df[name]

		elif type(column_name) == type(''):
			if column_name in self.df.columns:
				del self.df[column_name]

	def info(self):
		print(self.df.info())

	def get_data_as(self, data_type): #'DataFrame', 'read_csv', 'Tensor'
		if data_type == 'wData': return self
		if data_type == 'Tensor': 
			x = torch.from_numpy(self.df.values)
			x = Variable(x.type(self.torchDataType), requires_grad=False)
			X = x.to(self.device, non_blocking=True )
			return X
		if data_type == 'DataFrame': return self.df
		if data_type == 'ndarray': 
			#self.df.values[subset]
			return self.df.values
		if data_type == 'DataLoader':		# and self.torchloader is None 
			self.DM = wuml.DManager(self.df.values, self.Y)
			self.torchloader = DataLoader(dataset=self.DM, batch_size=self.batch_size, shuffle=True, pin_memory=True, num_workers=1)
			return self.torchloader

	def to_csv(self, path, add_row_indices=False, include_column_names=True, float_format='%.4f'):
		LCn = self.label_column_name

		if LCn is not None:
			self.df[LCn] = self.Y
			self.df.to_csv(path, index=add_row_indices, header=include_column_names, float_format=float_format)
			self.delete_column(LCn)
		elif self.Y is not None:
			self.df['label'] = self.Y
			self.df.to_csv(path, index=add_row_indices, header=include_column_names, float_format=float_format )
			self.delete_column('label')
		else:
			self.df.to_csv(path, index=add_row_indices, header=include_column_names, float_format=float_format)

	def __getitem__(self, item):
		if type(item).__name__ == 'str': return ensure_wData(self.df[item].to_frame())
		else: 
			return ensure_wData(self.df.iloc[item])

			#if return_data_as == 'DataFrame': return self.df.iloc[item]
			#if return_data_as == 'ndarray': return self.df.values[item]
			#if return_data_as == 'wData': return ensure_wData(self.df.iloc[item])


	def __str__(self): 
		return str(self.df)

	def __repr__(self): 
		return str(self.df)


	def plot_2_columns_as_scatter(self, column1, column2):
		X = self.df[column1].to_numpy()
		Y = self.df[column2].to_numpy()
		
		lp = wuml.scatter(figsize=(10,5))		# (width, height)
		lp.plot_scatter(X, Y, column1 + ' vs ' + column2, column1, column2, ticker_fontsize=8 )




import numpy as np
import wuml

class model_as_exponential: #Assuming that data is 1 dimension
	def __init__(self, X):
		X = np.squeeze(wuml.ensure_numpy(X))
		self.λ = 1/np.mean(X, axis=0)

	def __call__(self, x):
		try: x[x<0]=np.NAN
		except: pass

		λ = self.λ
		S = λ*np.exp(-λ*x)

		try: S[np.isnan(S)] = 0
		except: pass

		return S

	def cdf(self,x):
		λ = self.λ
		return (1 - np.exp(-λ*x))

		#[Area, error] = wuml.integrate(self.__call__, 0, x)
		#return Area




#!/usr/bin/env python

import numpy as np
import sys
import torch
from scipy.special import softmax
from torch.distributions.uniform import Uniform
from sklearn import preprocessing

np.set_printoptions(precision=4)
np.set_printoptions(threshold=sys.maxsize)
np.set_printoptions(linewidth=400)
np.set_printoptions(suppress=True)
torch.set_printoptions(edgeitems=3)
#torch.set_printoptions(threshold=10_00)
torch.set_printoptions(linewidth=400)
torch.set_printoptions(sci_mode=False)



def gumbel(pᵢ, τ=0.1, device='cpu'):		#The rows should add up to 1
	'''
		Notes : https://github.com/endsley/math_notebook/blob/master/neural_network/Gumbel_Softmax.pdf

		The gumbel softmax generates samples based on a categorial distribution
		that can be incorporated into a neural network. 
		Given a categorical distribution of {pᑊ pᒾ pᶾ ...}, it will generate one-hot vectors given these probabilities.

		Implement: Make sure that the rows add up to 1
	'''
	if type(pᵢ).__name__ == 'ndarray':	#The rows should add up to 1
		pᵢ = np.atleast_2d(pᵢ)
		logit = np.log(pᵢ)
	
		R = np.random.rand(logit.shape[0], logit.shape[1])
		ε = -np.log(-np.log(R))
		
		noisy_logits = (ε + logit) / τ
		C = softmax(noisy_logits, axis=1)
		return C

	elif type(pᵢ).__name__ == 'Tensor':	#The rows should add up to 1
		pᵢ = pᵢ.to(device, non_blocking=True)
		pᵢ += 0.00001			# Note: log(0) will give a NaN value so add noise
		logit = torch.log(pᵢ)

		uniform_dist = Uniform(1e-30, 1.0, )	
		uniform = uniform_dist.rsample(sample_shape=logit.size())
		uniform = uniform.to(device, non_blocking=True )
		ε = -torch.log(-torch.log(uniform))
	
		noisy_logits = (ε + logit) / τ
		C = torch.nn.Softmax(dim=-1)(noisy_logits)

		if torch.isnan(C).any():
			print('\n nan was detected inside gumbel\n')
			import pdb; pdb.set_trace()

		return C

if __name__ == "__main__": 
	X = np.array([[0.4405, 0.4045, 0.0754, 0.0796],			# The rows should add up to 1
					[0.2287, 0.2234, 0.2676, 0.2802],
					[0.2518, 0.2524, 0.1696, 0.3262],
					[0.2495, 0.1744, 0.2126, 0.3635],
					[0.1979, 0.31  , 0.2165, 0.2755],
					[0.2003, 0.2329, 0.2982, 0.2686]])
	
	# numpy implementation
	C = np.zeros((6,4))
	for n in range(10000):
		C += np.round(gumbel(X))
	
	C = C/10000
	
	# torch implementation
	Ct = torch.zeros(6,4)
	for n in range(10000):
		Ct += torch.round(gumbel(torch.from_numpy(X)))
	
	Ct = Ct.numpy()/10000

	print('True Probability')	
	print(X,'\n')

	print('Gumbel generated Probability from sampling')
	print(C,'\n')

	print('Gumbel generated Probability from sampling via pytorch')
	print(Ct,'\n')




#!/usr/bin/env python

from sklearn.model_selection import train_test_split
from sklearn.model_selection import KFold
from sklearn import preprocessing
import pandas as pd
import numpy as np
import sys
import os

#if os.path.exists('/home/chieh/code/wPlotLib'):
#	sys.path.insert(0,'/home/chieh/code/wPlotLib')

from wuml.IO import *
from wuml.data_stats import *
from wuml.data_loading import *
from wuml.type_check import *

np.set_printoptions(precision=4)
np.set_printoptions(threshold=30)
np.set_printoptions(linewidth=300)
np.set_printoptions(suppress=True)
np.set_printoptions(threshold=sys.maxsize)

#import pdb; pdb.set_trace()

def remove_rows_with_too_much_missing_entries(data, threshold=0.6, newDataFramePath=''):
	'''
		If a row is kept if it has more than "threshold" percentage of normal data
		data is a pandas format, it can be converted this way
		pd.DataFrame(data=data[1:,1:],    # values
						index=data[1:,0],    # 1st column as index
						columns=data[0,1:])  # 1st row as the column names
	'''
	df = ensure_DataFrame(data)

	X = df.values
	n = X.shape[0]
	d = X.shape[1]

	pth = './results/'
	ensure_path_exists('./results')
	#ensure_path_exists(pth)

	#	Obtain a decimated dataframe
	limitPer = int(d * threshold)
	df_decimated = df.dropna(thresh=limitPer, axis=0)
	oldID = df.index.values.tolist()
	newID = df_decimated.index.values.tolist()
	removed_samples = set(oldID).difference(newID)

	#	Record the results
	output_str = 'Original dataFrame dimension : %d samples,  %d dimensions\n'%(n, d)
	output_str += 'Deciated dataFrame dimension : %d samples,  %d dimensions\n'%(df_decimated.values.shape[0], df_decimated.values.shape[1])
	output_str += 'Removed Rows missing at least these percentage of entries : %.3f\n'%(1-threshold)

#	#	Record the removed features 
	output_str += '\nID of Removed Rows\n'
	output_str += str(removed_samples)

	write_to(output_str, pth + 'row_decimation_info.txt')
	if newDataFramePath != '': df_decimated.to_csv(path_or_buf=newDataFramePath, index=False)

	df_decimated = wData(dataFrame=df_decimated)
	return df_decimated


def remove_columns_with_too_much_missing_entries(data, threshold=0.6, newDataFramePath=''):
	'''
		If a column is kept if it has more than "threshold" percentage of normal data
		data is a pandas format, it can be converted this way
		pd.DataFrame(data=data[1:,1:],    # values
						index=data[1:,0],    # 1st column as index
						columns=data[0,1:])  # 1st row as the column names
	'''
	df = ensure_DataFrame(data)

	X = df.values
	n = X.shape[0]
	d = X.shape[1]

	pth = './results/'
	ensure_path_exists('./results')
	#ensure_path_exists(pth)

	#	Obtain a decimated dataframe
	limitPer = int(n * threshold)
	df_decimated = df.dropna(thresh=limitPer, axis=1)
	oldColumns = df.columns.values.tolist()
	newColumns = df_decimated.columns.values.tolist()
	removed_columns = set(oldColumns).difference(newColumns)

	#	Record the results
	output_str = 'Original data dimension : %d samples,  %d dimensions\n'%(n, d)
	output_str += 'Deciated data dimension : %d samples,  %d dimensions\n'%(df_decimated.values.shape[0], df_decimated.values.shape[1])
	output_str += 'Removed Columns missing at least these percentage of entries : %.3f\n'%(1-threshold)

	#	Record the removed features 
	output_str += '\nRemoved Columns + Missing Percentage'

	if len(removed_columns) > 0:
		max_width = len(max(removed_columns, key=len))
		for column in removed_columns:
			x = df[column].values
			missing_percentage = np.sum(np.isnan(x))/n
			output_str += (('\n\t%-' + str(max_width) + 's\t%.2f')%(column, missing_percentage))
	
		#	Record the retained features 
		output_str += '\n\nRetained Columns + Missing Percentage'
		for column in df_decimated:
			x = df_decimated[column].values
			missing_percentage = np.sum(np.isnan(x))/n
			output_str += (('\n\t%-' + str(max_width) + 's\t%.2f')%(column, missing_percentage))
	else:
		output_str += '\n\tNo columns were removed.'
		

	write_to(output_str, pth + 'column_decimation_info.txt')
	if newDataFramePath != '': df_decimated.to_csv(path_or_buf=newDataFramePath, index=False)

	return ensure_data_type(df_decimated, type_name=type(data).__name__)

def decimate_data_with_missing_entries(data, column_threshold=0.6, row_threshold=0.6,newDataFramePath=''):
	'''
		It will automatically remove rows and columns of a dataFrame with missing entries.
	'''

	dfo = ensure_DataFrame(data)
	dfSize = 'Data size:' + str(dfo.shape)

	mdp = np.array(identify_missing_data_per_feature(dfo))
	#x = np.arange(1, len(mdp)+1)
	colnames = dfo.columns.to_numpy()

	lp = wplotlib.bar(figsize=(10,5))
	lp.plot_bar(colnames, mdp, 'Before Missing Percentage', 'Feature ID', 'Percentage Missing', 
					imgText=dfSize, subplot=121, ylim=[0,1], 
					xticker_rotate=90)

	df = remove_columns_with_too_much_missing_entries(dfo, threshold=column_threshold)
	df_decimated = remove_rows_with_too_much_missing_entries(df, threshold=row_threshold, newDataFramePath=newDataFramePath)
	dfSize = 'Data size:' + str(df_decimated.shape)

	mdp = np.array(identify_missing_data_per_feature(df_decimated))
	#x = np.arange(1, len(mdp)+1)
	colnames = df.columns.to_numpy()
	lp.plot_bar(colnames, mdp, 'After Missing Percentage', 'Feature ID', 'Percentage Missing', 
					imgText=dfSize, subplot=122, ylim=[0,1], 
					xticker_rotate=90)
	lp.show()

	return df_decimated


def center_and_scale(wuData, return_type=None):
	X = ensure_numpy(wuData)
	X = preprocessing.scale(X)

	if type(wuData).__name__ == 'wData': 
		wuData.df = pd.DataFrame(data=X, columns=wuData.df.columns)
	elif type(wuData).__name__ == 'ndarray': 
		wuData.df = pd.DataFrame(data=X)
	elif type(wuData).__name__ == 'DataFrame': 
		wuData.df = pd.DataFrame(data=X, columns=wuData.columns)
	else:
		raise ValueError('Error: Cannot center data since %s is not a recongized data type.'%type(wuData).__name__)

	wuData.X = wuData.df.values

	if return_type is None:
		return ensure_data_type(wuData, type_name=type(wuData).__name__)
	else:
		return ensure_data_type(wuData, type_name=return_type)


def center_scale_with_missing_data(X, replace_nan_with_0=False): 
	'''
		For each column, find μ, σ while ignoring the entries that are zero. 
	'''
	d = X.shape[1]
	ignore_column_with_0_σ = []
	for i in range(d):
		x = X[:,i]
		ẋ = x[np.invert(np.isnan(x))]
		ẋ = ẋ - np.mean(ẋ)
		σ = np.std(ẋ)

		if σ < 0.00001:
			ignore_column_with_0_σ.append(i)
		else:
			X[np.invert(np.isnan(x)), i] = ẋ/σ

	for i in ignore_column_with_0_σ:
		X = np.delete(X, i , axis=1)	# delete column with σ=0

	if replace_nan_with_0:
		X = np.nan_to_num(X)

	return X, ignore_column_with_0_σ

def split_training_test(data, label=None, data_name=None, data_path=None, save_as='ndarray', test_percentage=0.1, xdata_type="%.4f", ydata_type="%d"):
	X = ensure_numpy(data)
	Y = None

	if label is not None: Y = label
	if type(data).__name__ == 'wData' and data.Y is not None: Y = data.Y
	if Y is None: raise ValueError('Error: The label Y is currently None, did you define it?')

	X_train, X_test, y_train, y_test = train_test_split(X, Y, test_size=test_percentage, random_state=42)

	if data_path is not None:
		Train_dat = data_path + data_name + '_train.csv'
		Train_label_dat = data_path + data_name + '_train_label.csv'
	
		Test_dat = data_path + data_name + '_test.csv'
		Test_label_dat = data_path + data_name + '_test_label.csv'
	
	
		if save_as == 'ndarray':
			np.savetxt(Train_dat, X_train, delimiter=',', fmt=xdata_type) 
			np.savetxt(Train_label_dat, y_train, delimiter=',', fmt=ydata_type) 
	
			np.savetxt(Test_dat, X_test, delimiter=',', fmt=xdata_type) 
			np.savetxt(Test_label_dat, y_test, delimiter=',', fmt=ydata_type) 
		elif save_as == 'DataFrame':
			XTrain_df = pd.DataFrame(data=X_train, columns=data.df.columns)
			XTest_df =  pd.DataFrame(data=X_test, columns=data.df.columns)
	
			XTrain_df['label'] = y_train
			XTest_df['label'] = y_test
	
			XTrain_df.to_csv(Train_dat, index=False, header=True)
			XTest_df.to_csv(Test_dat, index=False, header=True)

	X_train = ensure_wData(X_train, column_names=data.df.columns)
	X_train.Y = y_train

	X_test = ensure_wData(X_test, column_names=data.df.columns)
	X_test.Y = y_test

	return [X_train, X_test, y_train, y_test]

def gen_10_fold_data(data_name, data_path='./data/'):

	xpath = data_path + data_name

	X = np.loadtxt(xpath + '.csv', delimiter=',', dtype=np.float64)			
	Y = np.loadtxt(xpath + '_label.csv', delimiter=',', dtype=np.int32)			

	fold_path = xpath + '/'
	if os.path.exists(fold_path): 
		pass
	else:
		os.mkdir(fold_path)

	kf = KFold(n_splits=10, shuffle=True)
	kf.get_n_splits(X)
	loopObj = enumerate(kf.split(X))

	for count, data in loopObj:
		[train_index, test_index] = data

		X_train, X_test = X[train_index], X[test_index]
		Y_train, Y_test = Y[train_index], Y[test_index]
		
		np.savetxt( fold_path + data_name + '_' + str(count+1) + '.csv', X_train, delimiter=',', fmt='%.6f') 
		np.savetxt( fold_path + data_name + '_' + str(count+1) + '_label.csv', Y_train, delimiter=',', fmt='%d') 
		np.savetxt( fold_path + data_name + '_' + str(count+1) + '_test.csv', X_test, delimiter=',', fmt='%.6f') 
		np.savetxt( fold_path + data_name + '_' + str(count+1) + '_label_test.csv', Y_test, delimiter=',', fmt='%d') 

def rearrange_sample_to_same_class(X,Y):
	l = np.unique(Y)
	newX = np.empty((0, X.shape[1]))
	newY = np.empty((0))

	for i in l:
		indices = np.where(Y == i)[0]
		newX = np.vstack((newX, X[indices, :]))
		newY = np.hstack((newY, Y[indices]))

	return [newX, newY]


#	Relating all samples to the most likely one, with p(X1) > p(Xi) for all i
#	Given X1 X2 with p(X1)/p(X2)=2  the weight for X1 = 1, and X2 = 2
def get_likelihood_weight(data, weight_names=None):
	X = ensure_numpy(data)
	weight_names = ensure_list(weight_names)

	Pₓ = wuml.KDE(X)
	logLike = Pₓ(X, return_log_likelihood=True)
	max_likely = np.max(logLike)
	ratios = np.exp(max_likely - logLike)

	return wuml.wData(X_npArray=ratios, column_names=weight_names)

def map_data_between_0_and_1(data, output_type_name='wData', map_type='linear'): # map_type: linear, or cdf
	X = ensure_numpy(data)

	if map_type=='linear':
		min_max_scaler = preprocessing.MinMaxScaler()
		newX = min_max_scaler.fit_transform(X)
	else:
		n = X.shape[0]
		d = X.shape[1]
	
		newX = np.zeros(X.shape)
	
		for i in range(d):
			column = X[:,i]
			residual_dat = column[np.isnan(column) == False]
			minV = np.min(residual_dat) - 3
	
			if len(residual_dat) < 5:
				print('Error: column %d only has %d samples, you must at least have 6 samples for kde'%(i, len(residual_dat)))
				print('\nTry Removing this feature')
				sys.exit()
	
			if len(np.unique(residual_dat)) == 1:
				newX[:,i] = np.ones(n)
			else:
				Pₓ = wuml.KDE(residual_dat)
				for j, itm in enumerate(X[:,i]):
					if np.isnan(itm):
						newX[j,i] = np.nan
					else:
						newX[j,i] = Pₓ.integrate(minV, X[j,i])


	#	This ensures that the columns labels are copied correctly
	if type(data).__name__ == 'ndarray': 
		df = pd.DataFrame(newX)
	elif type(data).__name__ == 'wData': 
		df = pd.DataFrame(newX)
		df.columns = data.df.columns
	elif type(data).__name__ == 'DataFrame': 
		df = pd.DataFrame(newX)
		df.columns = data.columns
	elif type(data).__name__ == 'Tensor': 
		X = data.detach().cpu().numpy()
		df = pd.DataFrame(X)

	output = ensure_data_type(df, type_name=output_type_name)
	if output_type_name=='wData': 
		try: 
			output.Y = data.Y
			output.label_column_name = data.label_column_name
		except: pass

	return output


def use_cdf_to_map_data_between_0_and_1(data, output_type_name='wData'):
	print("This function is deprecated, instead use wuml.map_data_between_0_and_1(data, map_type='linear')")

	X = ensure_numpy(data)
	n = X.shape[0]
	d = X.shape[1]

	newX = np.zeros(X.shape)

	for i in range(d):
		column = X[:,i]
		residual_dat = column[np.isnan(column) == False]
		minV = np.min(residual_dat) - 3

		if len(residual_dat) < 5:
			print('Error: column %d only has %d samples, you must at least have 6 samples for kde'%(i, len(residual_dat)))
			print('\nTry Removing this feature')
			sys.exit()

		if len(np.unique(residual_dat)) == 1:
			newX[:,i] = np.ones(n)
		else:
			Pₓ = wuml.KDE(residual_dat)
			for j, itm in enumerate(X[:,i]):
				if np.isnan(itm):
					newX[j,i] = np.nan
				else:
					newX[j,i] = Pₓ.integrate(minV, X[j,i])


	#	This ensures that the columns labels are copied correctly
	if type(data).__name__ == 'ndarray': 
		df = pd.DataFrame(newX)
	elif type(data).__name__ == 'wData': 
		df = pd.DataFrame(newX)
		df.columns = data.df.columns
	elif type(data).__name__ == 'DataFrame': 
		df = pd.DataFrame(newX)
		df.columns = data.columns
	elif type(data).__name__ == 'Tensor': 
		X = data.detach().cpu().numpy()
		df = pd.DataFrame(X)

	output = ensure_data_type(df, type_name=output_type_name)
	if output_type_name=='wData': 
		try: 
			output.Y = data.Y
			output.label_column_name = data.label_column_name
		except: pass

	return output
